/*
 *   LADCCA 
 *    
 *   Copyright (C) 2002 Robert Ham (node@users.sourceforge.net)
 *    
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define _GNU_SOURCE

#include "ac-config.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>

#undef HAVE_CONFIG_H

#ifdef HAVE_LIBREADLINE
#  if defined(HAVE_READLINE_READLINE_H)
#    include <readline/readline.h>
#  elif defined(HAVE_READLINE_H)
#    include <readline.h>
#  else /* !defined(HAVE_READLINE_H) */
extern char *readline ();
#  endif /* !defined(HAVE_READLINE_H) */
char *cmdline = NULL;
#else /* !defined(HAVE_READLINE_READLINE_H) */
/* no readline */
#endif /* HAVE_LIBREADLINE */

#ifdef HAVE_READLINE_HISTORY
#  if defined(HAVE_READLINE_HISTORY_H)
#    include <readline/history.h>
#  elif defined(HAVE_HISTORY_H)
#    include <history.h>
#  else /* !defined(HAVE_HISTORY_H) */
extern void add_history ();
extern int write_history ();
extern int read_history ();
#  endif /* defined(HAVE_READLINE_HISTORY_H) */
/* no history */
#endif /* HAVE_READLINE_HISTORY */

#include <ladcca/ladcca.h>
#include <ladcca/debug.h>

#include "ladcca_control.h"


void
ladcca_control_init (ladcca_control_t * control)
{
  control->project = NULL;
  control->client = NULL;
  control->projects = NULL;
  
  rl_initialize ();
  rl_prep_terminal (1); 
}

void
ladcca_control_free (ladcca_control_t * control)
{
  if (control->project)
    free (control->project);
  
  rl_deprep_terminal ();
}

void
ladcca_control_set_project (ladcca_control_t * control, const char * project)
{
  if (control->project)
    free (control->project);
  
  if (project)
    control->project = cca_strdup (project);
  else
    control->project = NULL;
}

void
ladcca_control_set_client (ladcca_control_t * control, cca_client_t * client)
{
  control->client = client;
}

void
get_projects (ladcca_control_t * control)
{
  cca_event_t * event;
  char * project;
  char * projects;
  cca_list_t * list;
  int selected_project_exists = 0;

  event = cca_event_new ();
  
  /* request all the projects */
  cca_event_set_type (event, CCA_Projects);
  cca_send_event (control->client, event);
  
  /* wait for the server to respond */
  CCA_DEBUG ("waiting for server to respond");
  while (!cca_get_pending_event_count (control->client))
    {
      CCA_DEBUG ("  waiting");
      usleep (10000);
    }
  CCA_DEBUG ("  server response recieved");
  
  event = cca_get_event (control->client);
  
  if (cca_event_get_type (event) != CCA_Projects)
    {
      fprintf (stderr, "%s: something went wrong; we've been sent an event of type %d when we should have gotten a CCA_Projects\n",
               __FUNCTION__, cca_event_get_type (event));
      cca_event_destroy (event);
      return;
    }
  
  /* clear the existing projects */
  list = control->projects;
  while (list)
    {
      free (list->data);
      list = cca_list_remove (list, list->data);
    }
  control->projects = NULL;
  
  /* extract the projects */
  projects = (char *) cca_event_get_string (event);
  project = strtok (projects, ":");
  
  if (project)
    do
      {
        CCA_DEBUGARGS ("adding project '%s'", project);
        control->projects = cca_list_append (control->projects, cca_strdup (project));
        if (control->project)
          if (strcmp (project, control->project) == 0)
            {
              CCA_DEBUG ("selected project found");
              selected_project_exists = 1;
            }
      }
    while ((project = strtok (NULL, ":")));
    
  CCA_DEBUGARGS ("project '%s' found: %d", control->project, selected_project_exists);

  if (!control->projects)
    {
      if (control->project)
        ladcca_control_set_project (control, NULL);
    }
  else
    if (control->project) /* check it still exists */
     if (!selected_project_exists)
       ladcca_control_set_project (control, NULL);

  cca_event_destroy (event);
}

void
ladcca_control_connect (ladcca_control_t * control)
{
  get_projects (control);
  if (control->projects)
    {
      cca_event_t * event;
      
      ladcca_control_set_project (control, control->projects->data);
      
      event = cca_event_new_with_type (CCA_Project);
      cca_event_set_string (event, control->project);
      cca_send_event (control->client, event);
    }
  else
    ladcca_control_set_project (control, NULL);
}

static void
command_help ()
{
  printf ("\n");
  printf ("  list              List projects\n");
  printf ("  project <name>    Select the project with name <name>\n");
  printf ("  dir <dir>         Set the selected project's directory to <dir>\n");
  printf ("  name <name>       Set the selected project's name to <name>\n");
  printf ("  save              Save the selected project\n");
  printf ("  restore <dir>     Restore a project from <dir>\n");
  printf ("  tar  <tarball>    Pack the selected project into <tarball>\n");
  printf ("  untar <tarball>   Unpack and restore a project from <tarball>\n");
  printf ("  quit              Exit the program\n");
  printf ("\n");
}

static void
command_list (ladcca_control_t * control)
{
  cca_list_t * list;
  char * str;
  
  get_projects (control);
  
  list = control->projects;
  
  if (!list)
    {
      printf ("  no projects\n");
    }
  
  while (list)
    {
      str = (char *) list->data;
      printf ("  %s\n", str);
      
      list = list->next;
    }
}

static void
command_project (ladcca_control_t * control, char * arg)
{
  cca_list_t * list;
  char * str;
  
  if (!arg || !strlen(arg))
    {
      if (control->project)
        printf ("  selected project: %s\n", control->project);
      else
        printf ("  no project selected\n");
      return;
    }
  
  get_projects (control);
  
  list = control->projects;
  
  while (list)
    {
      str = (char *) list->data;
      if (strcmp (str, arg) == 0)
        {
          cca_event_t * event;

          ladcca_control_set_project (control, arg);

          event = cca_event_new ();
          cca_event_set_type (event, CCA_Project);
          cca_event_set_string (event, arg);
          cca_send_event (control->client, event);
          
          printf ("  project set to '%s'\n", arg);
          
          return;
        }
      
      list = list->next;
    }
  
  printf ("  unknown project '%s'\n", arg);
}

static void
command_dir (ladcca_control_t * control, char * arg)
{
  cca_event_t * event;
  
  if (!control->project)
    {
      printf ("  no project selected\n");
      return; 
    }
      
  event = cca_event_new ();
  cca_event_set_type (event, CCA_Project_Dir);
      
  if (!arg || !strlen(arg))
    { /* ask the server what the dir is */
      CCA_DEBUGARGS ("asking server for directory; event string: '%s'",
                     cca_event_get_string (event));
      cca_send_event (control->client, event);
      
      while (!cca_get_pending_event_count (control->client))
        usleep (10000);
      
      event = cca_get_event (control->client);
      if (cca_event_get_type (event) != CCA_Project_Dir)
        {
          fprintf (stderr, "%s: didn't recieve response to CCA_Project_Dir request\n", __FUNCTION__);
          cca_event_destroy (event);
          return;
        }
      
      printf ("  %s dir: '%s'\n", control->project, cca_event_get_string (event));
      cca_event_destroy (event);
      return;
    }
  
  cca_event_set_string (event, arg);
  cca_send_event (control->client, event);
  printf ("  dir set to '%s'\n", arg);
}

static void
command_restore (ladcca_control_t * control, char * arg)
{
  cca_event_t * event;
  
  if (!arg || !strlen(arg))
    {
      printf (" command requires an argument\n");
      return;
    }
  
  event = cca_event_new_with_type (CCA_Restore);
  cca_event_set_string (event, arg);
  cca_send_event (control->client, event);
  printf ("  restored project in dir '%s'\n", arg);
}

static void
command_name (ladcca_control_t * control, char * arg)
{
  cca_event_t * event;
  
  if (!control->project)
    {
      printf ("  no project selected\n");
      return; 
    }
      
  if (!arg || !strlen(arg))
    {
      printf (" command requires an argument\n");
      return;
    }
  
  event = cca_event_new ();
  cca_event_set_type (event, CCA_Project_Name);
  cca_event_set_string (event, arg);
  cca_send_event (control->client, event);
  
  ladcca_control_set_project (control, arg);
  printf ("  project name set to '%s'\n", arg);
  get_projects (control);
}

static void
command_save (ladcca_control_t * control)
{
  cca_event_t * event;
  
  if (!control->project)
    {
      printf ("  no project selected\n");
      return;
    }
  
  get_projects (control);
  
  if (!control->project) {
    printf ("  project disappeared\n");
    return;
  }
  
  event = cca_event_new_with_type (CCA_Save);
  cca_send_event (control->client, event);
  
/*  printf ("  saving: ");
  fflush (stdout);
  
  while (!saved)
    {
      while (!cca_get_pending_event_count (control->client))
        {
          usleep (10000);
        }
      
      event = cca_get_event (control->client);
      
      switch (cca_event_get_type (event))
      {
        case CCA_Save_Percentage:
          perc = cca_event_get_string(event)[0];
          for (i = perc_done + 1; i <= perc; i++)
            {
              if ( (int)(i / 10) * 10 == i)
                printf ("%d%%", i);
              else
                printf (".");
              fflush (stdout);
            }
          break;
        
        case CCA_Save_Complete:
          saved = 1;
          break;
      }
      
      cca_event_destroy (event);
    }
  
  printf ("\n");*/
}

static void
deal_with_command (ladcca_control_t * control, char * command)
{
  char * ptr;

  while (isspace (*command))
    command++;
  
  ptr = strchr (command, ' ');
  if (ptr)
    while (isspace (*ptr))
      ptr++;
    
  if (strncasecmp (command, "help", 1) == 0)
    {
      command_help ();
      return;
    }
  
  if (strncasecmp (command, "list", 1) == 0)
    {
      command_list (control);
      return;
    }
  
  if (strncasecmp (command, "project", 1) == 0)
    {
      command_project (control, ptr);
      return;
    }

  if (strncasecmp (command, "dir", 1) == 0)
    {
      command_dir (control, ptr);
      return;
    }

  if (strncasecmp (command, "name", 1) == 0)
    {
      command_name (control, ptr);
      return;
    }
  
  if (strncasecmp (command, "save", 1) == 0)
    {
      command_save (control);
      return;
    }

  if (strncasecmp (command, "restore", 1) == 0)
    {
      command_restore (control, ptr);
      return;
    }
  
  ptr = strchr (command, ' ');
  if (ptr)
    *ptr = '\0';
  printf ("unknown command '%s'\n", command);
  if (ptr)
    *ptr = ' ';
}

void ladcca_control_main (ladcca_control_t * control)
{
  char * prompt;
  size_t prompt_size = 64;
  size_t project_size;
  char * command;
  
  prompt = cca_malloc (prompt_size);
  
  printf ("\nLADCCA Control %s\n", PACKAGE_VERSION);
  printf ("\nconnected to server %s\n\n", cca_get_server_name (control->client));
  
  for (;;)
    {
      project_size = control->project ? strlen (control->project) : 0;
      
      if (project_size)
        {
          if (project_size + 4 > prompt_size)
            {
              free (prompt);
              prompt_size = project_size + 4;
              prompt = cca_malloc (prompt_size);
            }
          
          sprintf (prompt, "%s > ", control->project);
        }
      else
        {
          sprintf (prompt, "(no project) > ");
        }
      
      command = readline (prompt);
      
      if (!command)
        {
          printf("\n");
          exit (0);
        }
      
      if (strlen (command))
        {
          deal_with_command (control, command);
          
#ifdef HAVE_READLINE_HISTORY
          add_history (command);
#endif

        }
      
      free (command);
    }
  
}


/* EOF */


