/*
 *   LADCCA
 *    
 *   Copyright (C) 2002 Robert Ham (node@users.sourceforge.net)
 *    
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <ladcca/ladcca.h>
#include <ladcca/internal_headers.h>

#include "jack_patch.h"
#include "jack_mgr_client.h"
#include "jack_mgr.h"

void
jack_patch_init (jack_patch_t * patch)
{
  patch->from = NULL;
  patch->to = NULL;
  uuid_clear (patch->from_id);
  uuid_clear (patch->to_id);
}

void
jack_patch_free (jack_patch_t * patch)
{
  if (patch->from)
    free (patch->from);
    
  if (patch->to)
    free (patch->to);
  
  jack_patch_init (patch);
}

jack_patch_t *
jack_patch_new ()
{
  jack_patch_t * patch;
  patch = cca_malloc (sizeof (jack_patch_t));
  jack_patch_init (patch);
  return patch;
}

jack_patch_t *
jack_patch_dup (const jack_patch_t * other)
{
  jack_patch_t * patch;
  uuid_t id;
  
  patch = jack_patch_new ();
  
  jack_patch_set_from (patch, jack_patch_get_from (other));
  jack_patch_set_to   (patch, jack_patch_get_to (other));
  
  jack_patch_get_from_id (other, id);
  jack_patch_set_from_id (patch, id);

  jack_patch_get_to_id (other, id);
  jack_patch_set_to_id (patch, id);
  
  return patch;
}


void
jack_patch_destroy (jack_patch_t * patch)
{
  jack_patch_free (patch);
  free (patch);
}

const char *
jack_patch_get_from (const jack_patch_t * patch)
{
  return patch->from;
}

const char *
jack_patch_get_to   (const jack_patch_t * patch)
{
  return patch->to;
}


const char *
jack_patch_get_from_port (const jack_patch_t * patch)
{
  return get_jack_port_name_only (patch->from);
}

const char *
jack_patch_get_to_port   (const jack_patch_t * patch)
{
  return get_jack_port_name_only (patch->to);
}

const char *
jack_patch_get_from_client (const jack_patch_t * patch)
{
  return get_jack_client_name_only (patch->from);
}

const char *
jack_patch_get_to_client   (const jack_patch_t * patch)
{
  return get_jack_client_name_only (patch->to);
}


void
jack_patch_set_from (jack_patch_t * patch, const char * from)
{
  set_string_property (patch->from, from);
}

void
jack_patch_set_to (jack_patch_t * patch, const char * to)
{
  set_string_property (patch->to, to);
}

void
jack_patch_get_from_id (const jack_patch_t * patch, uuid_t id)
{
  uuid_copy (id, ((jack_patch_t *) patch)->from_id);
}

void
jack_patch_set_from_id (jack_patch_t * patch, uuid_t id)
{
  uuid_copy (patch->from_id, id);
}

void
jack_patch_get_to_id (const jack_patch_t * patch, uuid_t id)
{
  uuid_copy (id, ((jack_patch_t *) patch)->to_id);
}

void
jack_patch_set_to_id (jack_patch_t * patch, uuid_t id)
{
  uuid_copy (patch->to_id, id);
}

void
jack_patch_set_spec (jack_patch_t * patch, const char * string)
{
  char * str, * ptr;
  
  str = cca_strdup (string);
  
  ptr = strchr (str, '/');
  if (!ptr)
    {
      free (str);
      return;
    }
  
  *ptr = '\0';  
  ptr++;
  
  jack_patch_set_from (patch, str);
  jack_patch_set_to   (patch, ptr);
  
  free (str);
}

void
jack_patch_set_port_name (cca_list_t * jack_mgr_clients, char ** port_name_ptr, uuid_t id)
{
  cca_list_t * list;
  char * client_name;
  char * ptr;
  jack_mgr_client_t * client = NULL;
  int found = 0;
  
  if (!*port_name_ptr)
    return;

  /* extract the client name */
  client_name = cca_strdup (*port_name_ptr);
  ptr = strchr (client_name, ':');
  *ptr = '\0';
  
  /* find the client */
  for (list = jack_mgr_clients; list; list = cca_list_next (list))
    {
      client = (jack_mgr_client_t *) list->data;
      
      jack_mgr_client_lock (client);
      
      if (strcmp (jack_mgr_client_get_name (client), client_name) == 0)
        {
          found = 1;
          break;
        }
      
      jack_mgr_client_unlock (client);
    }
  
  if (!found)
    return;
  
  if (found)
    {
      char id_str[37];
      
      jack_mgr_client_get_id (client, id);
      uuid_unparse (id, id_str);
      
      free (client_name);
      client_name = strchr (*port_name_ptr, ':') + 1;
      ptr = cca_malloc (sizeof (id_str) + 1 + strlen (client_name) + 1);
      sprintf (ptr, "%s:%s", id_str, client_name);
      
      free (*port_name_ptr);
      *port_name_ptr = ptr;
    }

  jack_mgr_client_unlock (client);
}

void
jack_patch_set (jack_patch_t * patch, cca_list_t * jack_mgr_clients)
{
  jack_patch_set_port_name (jack_mgr_clients, &patch->from, patch->from_id);
  jack_patch_set_port_name (jack_mgr_clients, &patch->to, patch->to_id);
}

int
jack_patch_unset_port_name (cca_list_t * jack_mgr_clients, char ** port_name_ptr)
{
  const cca_list_t * list;
  jack_mgr_client_t * client = NULL;
  int found = 0;
  uuid_t port_id, client_id;
  char * port_name;
  char * ptr;
  int err;
  char * new_port_name;
  
  if (!*port_name_ptr)
    return 1;
  
  port_name = cca_strdup (*port_name_ptr);
  ptr = strchr (port_name, ':');
  *ptr = '\0';
  
  err = uuid_parse (port_name, port_id);
  if (err == -1)
    {
      free (port_name);
      return 0;
    }

  /* find the client */
  for (list = jack_mgr_clients; list; list = cca_list_next (list))
    {
      client = (jack_mgr_client_t *) list->data;
      
      jack_mgr_client_lock (client);
      jack_mgr_client_get_id (client, client_id);
      
      if (uuid_compare (port_id, client_id) == 0) 
        {
          found = 1;
          break;
        }

      jack_mgr_client_unlock (client);
    }
    
  if (!found)
    return 1;
  
  ptr = strchr (port_name, '\0') + 1;
  new_port_name = cca_malloc (strlen (ptr) + 1 + strlen (jack_mgr_client_get_name (client)) + 1);
  sprintf (new_port_name, "%s:%s", jack_mgr_client_get_name (client), ptr);
  free (*port_name_ptr);
  *port_name_ptr = new_port_name;
  
  jack_mgr_client_unlock (client);
  
  free (port_name);
  
  return 0;
}

int
jack_patch_unset (jack_patch_t * patch, cca_list_t * jack_mgr_clients)
{
  int err;
  
  err = jack_patch_unset_port_name (jack_mgr_clients, &patch->from);
  if (err)
    return err;
  
  err = jack_patch_unset_port_name (jack_mgr_clients, &patch->to);
  if (err)
    return err;
  
  return 0;
}


const char *
jack_patch_get_spec (jack_patch_t * patch)
{
  static char * spec = NULL;
  static size_t spec_size = 0;
  size_t patch_spec_size;
  
  patch_spec_size = strlen (jack_patch_get_from (patch)) + 1 +
                    strlen (jack_patch_get_to   (patch)) + 1;
  
  if (patch_spec_size > spec_size)
    {
      spec_size = patch_spec_size;
      spec = cca_realloc (spec, spec_size);
    }
  
  sprintf (spec, "%s/%s", jack_patch_get_from (patch), jack_patch_get_to (patch));
  
  return spec;
}

void
jack_patch_switch_clients (jack_patch_t * patch)
{
  char * to;
  uuid_t id;
  
  to          = patch->to;
  patch->to   = patch->from;
  patch->from = to;

  uuid_copy (id, patch->to_id);
  uuid_copy (patch->to_id, patch->from_id);
  uuid_copy (patch->from_id, id);
}

/* EOF */

