/*
 *   LADCCA
 *    
 *   Copyright (C) 2002 Robert Ham (node@users.sourceforge.net)
 *    
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <sys/types.h>
#include <sys/socket.h>
#include <string.h>
#include <arpa/inet.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
                     
#include <ladcca/ladcca.h>
#include <ladcca/internal_headers.h>

#include "loader.h"

#define XTERM_COMMAND_EXTENSION "&& sh || sh"

void
loader_init (loader_t * loader)
{
  loader->server_socket  = 0;
  loader->loader_socket  = 0;
  loader->loader_pid     = -1;
  loader->load_in_terms  = 0;
}

loader_t *
loader_new ()
{
  loader_t * loader;
  loader = cca_malloc (sizeof (loader_t));
  loader_init (loader);
  return loader;
}

loader_t *
loader_create (int load_in_terms)
{
  loader_t * loader;
  int sockets[2];
  int err;
  
  err = socketpair (AF_UNIX, SOCK_STREAM, 0, sockets);
  if (err == -1)
    {
      fprintf (stderr, "%s: could not create unix socket pair: %s\n",
               __FUNCTION__, strerror (errno));
      return NULL;
    }
  
  loader = loader_new ();
  
  loader->server_socket = sockets[0];
  loader->loader_socket = sockets[1];
  
  loader->load_in_terms = load_in_terms;
  
  return loader;
}

/* all args are enclosed in quotes */
static size_t
get_command_size (int argc, char ** argv)
{
  size_t size = 0;
  int i;
  
  for (i = 0; i < argc; i++)
    size += strlen (argv[i]) + 3;
  
  size += strlen (XTERM_COMMAND_EXTENSION);
  
  return size + 1;
}

static void
create_command (char * command_buffer, int argc, char ** argv)
{
  char * ptr;
  int i;
  
  ptr = command_buffer;

  for (i = 0; i < argc; i++)
    {
      sprintf (ptr, "\"%s\" ", argv[i]);
      ptr += strlen (ptr);
    }
  
  sprintf (ptr, "%s", XTERM_COMMAND_EXTENSION);
}

static void
loader_exec_program_in_xterm (int argc, char ** argv)
{
  size_t command_size;
  char * command_buffer;
  char * xterm_argv[6];
  
  command_size = get_command_size (argc, argv);
  
  command_buffer = cca_malloc (command_size);
  
  create_command (command_buffer, argc, argv);
  
  xterm_argv[0] = "xterm";
  xterm_argv[1] = "-e";
  xterm_argv[2] = "bash";
  xterm_argv[3] = "-c";
  xterm_argv[4] = command_buffer;
  xterm_argv[5] = NULL;

  /* execute it */
  execvp ("xterm", xterm_argv);
  
  fprintf (stderr, "%s: execing program '%s' in an xterm failed: %s\n",
           __FUNCTION__, command_buffer, strerror (errno));
  
  exit (1);
  
  
}

void
loader_exec_program (loader_t * loader, cca_client_t * client)
{
  int err;
  char * project_opt;
  char * server_opt;
  char port_opt[32];
  char id_opt[12+36+1];
  const char * program;
  char ** argv;
  int argc = 0;
  int i;
  
  program = cca_client_get_argv (client)[0];
  
  /* change the working dir */
  err = chdir (cca_client_get_working_dir (client));
  if (err == -1)
    {
      fprintf (stderr, "%s: could not change directory to working dir '%s' for program '%s': %s\n",
               __FUNCTION__, cca_client_get_working_dir (client), program, strerror (errno));
    }
  
  /* construct the args */
  
  
  project_opt = cca_malloc (strlen (cca_client_get_project_name (client)) + 18);
  sprintf (project_opt, "--ladcca-project=%s", cca_client_get_project_name (client));
  
  server_opt = cca_malloc (strlen (cca_client_get_server_name (client)) + 17);
  sprintf (server_opt, "--ladcca-server=%s", cca_client_get_server_name (client));
  
  sprintf (port_opt, "--ladcca-port=%d", cca_client_get_server_connected (client));
  
  sprintf (id_opt, "--ladcca-id=%s", cca_client_get_id_str (client));
  
  argc += cca_client_get_argc (client) + 4;
  argv = cca_malloc (sizeof (char *) * (argc + 1));
  
  for (i = 0; i < cca_client_get_argc (client); i++)
    argv[i] = cca_client_get_argv(client)[i];
  
  argv[(i++)] = project_opt;
  argv[(i++)] = server_opt;
  argv[(i++)] = port_opt;
  argv[(i++)] = id_opt;
  argv[i]   = NULL;

#ifdef LADCCA_DEBUG
  {
    int l;
    printf ("%s: executing command: ", __FUNCTION__);
    for (l = 0; l < argc; l++)
      {
        printf ("%s ", argv[l]);
      }
    putchar ('\n');
  }
#endif

  if (CCA_CLIENT_RUNS_IN_TERMINAL (client))
    loader_exec_program_in_xterm (argc, argv);

  /* execute it */
  execvp (program, argv);
  
  fprintf (stderr, "%s: execing program '%s' failed: %s\n",
           __FUNCTION__, program, strerror (errno));
  
  exit (1);
}

void
loader_exec (loader_t * loader, cca_client_t * client)
{
  pid_t pid;
  const char * program;
  
  program = cca_client_get_argv(client)[0];
  
  pid = fork ();
  switch (pid)
    {
    case -1:
      fprintf (stderr, "%s: could not fork in order to exec program '%s': %s",
               __FUNCTION__, program, strerror (errno));
      break;
    
    case 0:
      loader_exec_program (loader, client);
    
    default:
      CCA_DEBUGARGS ("forked to run program '%s'", program);
      break;
    }
}

void
loader_event (loader_t * loader, cca_comm_event_t * event)
{
  switch (cca_comm_event_get_type (event))
    {
    case CCA_Comm_Event_Close:
      CCA_DEBUG ("recieved close event, exiting")
      exit (0);
    case CCA_Comm_Event_Connect:
      loader_exec (loader, cca_comm_event_take_client (event));
      break;
    default:
      fprintf (stderr, "%s: recieved unknown event of type %d\n", __FUNCTION__, cca_comm_event_get_type (event));
      break;
    }
}

void
loader_run (loader_t * loader)
{
  cca_comm_event_t * event;
  int err;
  
  signal (SIGTERM, SIG_IGN);
  signal (SIGINT, SIG_IGN);
  signal (SIGHUP, SIG_IGN);
  
  /* don't need the server socket */
  err = close (loader->server_socket);
  if (err == -1)
    {
      fprintf (stderr, "%s: could not close server socket: %s\n",
               __FUNCTION__, strerror (errno));
    }
  
  /* apps don't need the loader socket */
  err = fcntl (loader->loader_socket, F_SETFD, FD_CLOEXEC);
  if (err == -1)
    {
      fprintf (stderr, "%s: could not set close-on-exec on loader socket: %s\n",
               __FUNCTION__, strerror (errno));
    }
  
  /* we don't care if apps finish */
  signal (SIGCHLD, SIG_IGN);
  
  for (;;)
    {
      CCA_DEBUG ("listening for event");
      event = cca_comm_event_new ();
      err = cca_comm_recv_event (loader->loader_socket, event);
      
      if (err == -2)
        {
          fprintf (stderr, "%s: server closed socket; exiting\n", __FUNCTION__);
          exit (1);
        }
      
      if (err == -1)
        {
          fprintf (stderr, "%s: error recieving event; exiting\n", __FUNCTION__);
          exit (1);
        }
      
      loader_event (loader, event);
      
      cca_comm_event_destroy (event);
    }
}

int
loader_fork (loader_t * loader)
{
  pid_t pid;
  int err;
  
  pid = fork ();
  
  switch (pid)
    {
    case -1:
      fprintf (stderr, "%s: error while forking: %s\n", __FUNCTION__, strerror (errno));
      return 1;
      
    case 0:
      loader_run (loader);
      exit (0);
      break;
      
    default:
      loader->loader_pid = pid;
      err = close (loader->loader_socket);
      if (err == -1)
        {
          fprintf (stderr, "%s: error closing loader socket: %s\n",
                   __FUNCTION__, strerror (errno));
        }
      break;
    }
  
  return 0;
}

void
loader_load (loader_t * loader, const cca_client_t * client)
{
  cca_comm_event_t event;
  int err;
  
  cca_comm_event_set_connect (&event, (cca_client_t *) client);
  
  err = cca_comm_send_event (loader->server_socket, &event);
  if (err < 0)
    {
      fprintf (stderr, "%s: error sending event to the loader\n", __FUNCTION__);
    }
}

void
loader_close (loader_t * loader)
{
  cca_comm_event_t event;
  int err;
  
  CCA_DEBUG ("sending close event to loader");
  cca_comm_event_set_type (&event, CCA_Comm_Event_Close);
  
  err = cca_comm_send_event (loader->server_socket, &event);
  if (err < 0)
    {
      fprintf (stderr, "%s: error sending Close event to loader; a dangling child process is likely\n", __FUNCTION__);
    }
  
  CCA_DEBUG ("close event sent");
}

void
loader_destroy (loader_t * loader)
{
  int err;
  
  err = close (loader->server_socket);
  if (err == -1)
    {
      fprintf (stderr, "%s: error closing loader's server-side socket: %s\n",
              __FUNCTION__, strerror (errno));
    }
  
  free (loader);
}

/* EOF */

