/*
 *   LADCCA
 *    
 *   Copyright (C) 2002 Robert Ham (node@users.sourceforge.net)
 *    
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define _GNU_SOURCE

#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <assert.h>
#include <limits.h>

#include <jack/jack.h>
#include <alsa/asoundlib.h>

#include <ladcca/ladcca.h>
#include <ladcca/internal_headers.h>

#include "project.h"
#include "client.h"
#include "store.h"
#include "jack_patch.h"
#include "alsa_patch.h"
#include "globals.h"
#include "server.h"
#include "jack_mgr.h"

#define ID_DIR      ".id"
#define CONFIG_DIR  ".config"
#define INFO_FILE   ".ladcca_info"

static void
project_init (project_t * project)
{
  project->name         = NULL;
  project->directory    = NULL;
  project->clients      = NULL;
  project->lost_clients = NULL;
}

project_t *
project_new (server_t * server)
{
  project_t * project;
  project = cca_malloc (sizeof (project_t));
  project_init (project);

  project->server = server;

  return project;
}

static void
project_free (project_t * project)
{
  cca_list_t * list;
  client_t * client;
  
  project_set_name (project, NULL);
  project_set_directory (project, NULL);
  
  list = project->clients;
  while (list)
    {
      client = (client_t *) list->data;
      
      client_destroy (client);
      
      list = list->next;
    }
  
  cca_list_free (project->clients);
  project->clients = NULL;
}

void
project_destroy (project_t * project)
{
  project_free (project);
  free (project);
}

const char *
project_get_name (const project_t * project)
{
  return project->name;
} 

const char *
project_get_directory (const project_t * project)
{
  return project->directory;
}

server_t *
project_get_server (project_t * project)
{
  return project->server;
}



void
project_set_directory (project_t * project, const char * directory)
{
  if (project->directory)
    free (project->directory);
  
  if (directory)
    {
      char * ptr;
      project->directory = cca_strdup (directory);
      ptr = project->directory + strlen (project->directory) - 1;
      if (*ptr == '/') *ptr = '\0';
      
      if (!cca_dir_exists (directory))
        cca_create_dir (directory);
    }
  else
    project->directory = NULL;
}

void
project_set_name (project_t * project, const char * name)
{
  cca_list_t * list;
  client_t * client;
  

  if (project->name)
    free (project->name);
  
  if (name)
    project->name = cca_strdup (name);
  else
    project->name = NULL;
  
  list = project->clients;
  while (list)
    {
      client = (client_t *) list->data;
      cca_client_set_project_name (client_get_cca_client (client), name);
      list = list->next;
    }
}

int 
project_name_exists (cca_list_t * projects, const char * name)
{
  project_t * project;

  while (projects)
    {
      project = (project_t *) projects->data;
      if (strcmp (name, project->name) == 0)
        return 1;
      
      projects = projects->next;
    }
  
  return 0;
}

client_t *
project_get_lost_client_by_id (project_t * project, uuid_t id)
{
  cca_list_t * list;
  client_t * client;
  uuid_t client_id;
  
  list = project->lost_clients;
  while (list)
    {
      client = (client_t *) list->data;
      
      cca_client_get_id (client_get_cca_client (client), client_id);
      
      if (uuid_compare (client_id, id) == 0)
        return client;
      
      list = list->next;
    }
  
  return NULL;
}

/**************************************
 ************ store stuff *************
 **************************************/

 
const char *
project_get_client_id_dir (project_t * project, client_t * client)
{
  static char * dir = NULL;
  
  if (dir)
    free (dir);
    
  dir = cca_strdup (cca_get_fqn ( cca_get_fqn (project_get_directory (project), ID_DIR),
                                  cca_client_get_id_str (client_get_cca_client (client)) ));
  
  return dir;
}

const char *
project_get_client_config_dir (project_t * project, client_t * client)
{
/*  dir = cca_strdup (cca_get_fqn (project_get_client_id_dir (project, client), CONFIG_DIR)); */
  get_store_and_return_fqn (project_get_client_id_dir (project, client), CONFIG_DIR);
}

const char *
project_get_client_name_dir (project_t * project, client_t * client)
{
  static char * dir = NULL;
  
  if (dir)
    free (dir);
    
  dir = cca_strdup (cca_get_fqn (project_get_directory (project), client_get_name (client)));
  
  return dir;
}

void
project_new_client (project_t * project, client_t * client)
{
  uuid_t id;
  cca_client_t * cca_client;
  
  cca_client = client_get_cca_client (client);
  
  cca_client_get_id (cca_client, id);
  if (uuid_is_null (id))
    client_generate_id (client);
  
  CCA_DEBUGARGS ("client on connection %ld now has id '%s'",
                 client_get_conn_id (client),
                 cca_client_get_id_str (cca_client));
  
  if (CCA_CLIENT_CONFIGURED_WITH_FILE (cca_client) ||
      CCA_CLIENT_CONFIGURED_WITH_DATA_SET (cca_client))
    cca_create_dir (project_get_client_id_dir (project, client));
  
  if (CCA_CLIENT_CONFIGURED_WITH_DATA_SET (client_get_cca_client (client)))
    client_store_open (client, project_get_client_config_dir (project, client));
  
  project->clients = cca_list_append (project->clients, client);

  CCA_DEBUGARGS ("added client '%s' to project '%s'",
                 cca_client_get_id_str (cca_client),
                 project_get_name (project));
}

void
project_name_client (project_t * project, client_t * client, const char * name)
{
  cca_list_t * list;
  client_t * p_client;
  cca_client_t * cca_client;
  int err;
  int linked = 0;
  
  cca_client = client_get_cca_client (client);
  
  CCA_DEBUGARGS ("naming client '%s' with name '%s'",
                 cca_client_get_id_str (cca_client),
                 name);
  
  /* check the client doesn't already have the name */
  if (client_get_name (client) && strcmp (name, client_get_name (client)) == 0)
    {
      CCA_DEBUGARGS ("client '%s' already has name '%s'; not setting", cca_client_get_id_str (cca_client), name);
      return;
    }
  
  /* check the name doesn't already exist */
  list = project->clients;
  while (list)
    {
      p_client = (client_t *) list->data;
      
      if (client_get_name (p_client) &&
          strcmp (client_get_name (p_client), name) == 0)
        {
          fprintf (stderr, "%s: a client '%s' already has the name '%s'; not setting name for client '%s'\n",
                   __FUNCTION__, 
                   cca_client_get_id_str (client_get_cca_client (p_client)),
                   name,
                   cca_client_get_id_str (cca_client));
          return;
        }
        
      list = list->next;
    }
  
  if (client_get_name (client) &&
      (CCA_CLIENT_CONFIGURED_WITH_DATA_SET (cca_client) ||
       CCA_CLIENT_CONFIGURED_WITH_FILE (cca_client)))
    { /* link should already exist */
      char * old_link;

      old_link = cca_strdup (project_get_client_name_dir (project, client));
      
      client_set_name (client, name);
      
      CCA_DEBUGARGS ("moving old link '%s' to '%s'", old_link, project_get_client_name_dir (project, client));

      err = rename (old_link, project_get_client_name_dir (project, client));
      if (err == -1)
        {
          fprintf (stderr, "%s: could not rename name link for client '%s': %s\n",
                   __FUNCTION__, cca_client_get_id_str (client_get_cca_client (client)),
                   strerror (errno));
        }
      else
        linked = 1;
      
      free (old_link);
    }
  
  if (!linked &&
      (CCA_CLIENT_CONFIGURED_WITH_DATA_SET (cca_client) ||
       CCA_CLIENT_CONFIGURED_WITH_FILE (cca_client)))
    {
      char * id_dir;
      
      client_set_name (client, name);
      id_dir = cca_strdup (project_get_client_id_dir (project, client));
      
      CCA_DEBUGARGS ("linking id dir '%s' to name dir '%s", id_dir,
                     project_get_client_name_dir (project, client));
                     
      err = symlink (id_dir, project_get_client_name_dir (project, client));
      if (err == -1)
         fprintf (stderr, "%s: could not create name link for client '%s': %s\n",
                  __FUNCTION__, cca_client_get_id_str (client_get_cca_client(client)), strerror (errno));
      
      free (id_dir);
    }
    
  client_set_name (client, name);
}

void
project_move_client (project_t * project, client_t * client, const char * new_project_dir)
{
  char * project_dir;
  cca_client_t * cca_client;
  int err;

  if (strcmp (project_get_directory (project), new_project_dir) == 0)
    return;


  project_dir = cca_strdup (project_get_directory (project));
  cca_client  = client_get_cca_client (client);

  if (CCA_CLIENT_CONFIGURED_WITH_DATA_SET (cca_client) ||
      CCA_CLIENT_CONFIGURED_WITH_FILE     (cca_client))  
    {
      /* move the client dir */
      char * old_dir;
      char * new_dir;
      
      old_dir = cca_strdup (project_get_client_id_dir (project, client));
      
      project_set_directory (project, new_project_dir);
      new_dir = (char *) project_get_client_id_dir (project, client);
      
      err = rename (old_dir, new_dir);
      if (err == -1)
        {
          fprintf (stderr, "%s: could not move id dir '%s' to new dir '%s': %s\n",
                   __FUNCTION__, old_dir, new_dir, strerror (errno));
          free (old_dir);
          return;
        }
      free (old_dir);
      
      /* make sure the store is looking at the new place */
      new_dir = cca_strdup (project_get_client_config_dir (project, client));
      store_set_dir (client_get_store (client), new_dir);  
      free (new_dir);
      
      /* possibly sort out the link */
      if (client_get_name (client))   
        {
          project_set_directory (project, project_dir);
          old_dir = (char *) project_get_client_name_dir (project, client);
          err = unlink (old_dir);
          if (err == -1)
            {
              fprintf (stderr, "%s: could not remove old name link '%s': %s\n",
                       __FUNCTION__, old_dir, strerror (errno));
            }
          

          old_dir = cca_strdup (project_get_client_id_dir (project, client));
          project_set_directory (project, new_project_dir);
          new_dir = (char *) project_get_client_name_dir (project, client);

          err = symlink (old_dir, new_dir);
          if (err == -1)
            {
              fprintf (stderr, "%s: could not create name link '%s' for client '%s': %s\n",
                       __FUNCTION__, new_dir, cca_client_get_id_str (cca_client), strerror (errno));
            }
 	  free (old_dir);
        }
    }
  
  project_set_directory (project, project_dir);
  free (project_dir);
}

void
project_move (project_t * project, const char * new_dir)
{
  cca_list_t * list;
  client_t * client;
  
  /* create the new dir */
  cca_create_dir (new_dir);
  
  /* move all the clients */
  list = project->clients;
  while (list)
    {
      client = (client_t *) list->data;
      
      project_move_client (project, client, new_dir);
      
      list = list->next;
    }
  
  /* move the project */
  /* FIXME */
}

void
project_restore_data_set (project_t * project, client_t * client)
{
  cca_client_t * cca_client;
  cca_client = client_get_cca_client (client);
  
  if (CCA_CLIENT_CONFIGURED_WITH_DATA_SET (cca_client))
    {
      char * key;
      cca_config_t * config;
      cca_list_t * list;
      cca_event_t * event;
      
      CCA_DEBUGARGS ("restoring data set for client '%s'", cca_client_get_id_str (cca_client));

      list = store_get_keys (client_get_store (client));

      /* send the event to the client */
      if (list)
        {
          event = cca_event_new_with_type (CCA_Restore_Data_Set);
          conn_mgr_send_client_cca_event (project->server->conn_mgr,
                                          client_get_conn_id (client), event);
        }
      
      for (; list; list = cca_list_next (list))
        {
          key = (char *) list->data;
          
          config = store_get_config (client_get_store (client), key);
          
          if (config)
            conn_mgr_send_client_cca_config (project->server->conn_mgr,
                                             client_get_conn_id (client),
                                             config);
        }
    }
}

void
project_resume_client (project_t * project, client_t * client, client_t * lost_client)
{
  cca_client_t * cca_client;
  cca_client_t * lost_cca_client;
  cca_event_t * event;
  uuid_t id;
  int err;
  

  /*
   * get all the necessary data from the lost client
   */
       cca_client = client_get_cca_client (client);
  lost_cca_client = client_get_cca_client (lost_client);
  
  client_set_name (client, client_get_name (lost_client));
  client->alsa_patches = lost_client->alsa_patches;
  client->jack_patches = lost_client->jack_patches;
      
  cca_client_get_id (lost_cca_client, id);
  cca_client_set_id (cca_client, id);
  cca_client_set_flags (cca_client, cca_client_get_flags (lost_cca_client));
  
  
  /*
   * kill the lost client
   */
  project->lost_clients = cca_list_remove (project->lost_clients, lost_client);
  client_destroy (lost_client);
  
  CCA_DEBUGARGS ("resuming client '%s'", cca_client_get_id_str (cca_client));
  
  /*
   * resume the client
   */
  if (CCA_CLIENT_CONFIGURED_WITH_FILE (cca_client)
      || CCA_CLIENT_CONFIGURED_WITH_DATA_SET (cca_client))
    cca_create_dir (project_get_client_id_dir (project, client));
  
  /* tell the client to load its files */
  if (CCA_CLIENT_CONFIGURED_WITH_FILE (cca_client))
    {
      event = cca_event_new_with_type (CCA_Restore_File);
      cca_event_set_string (event, project_get_client_id_dir (project, client));
      conn_mgr_send_client_cca_event (project->server->conn_mgr,
                                      client_get_conn_id (client), event);
    }

  if (CCA_CLIENT_CONFIGURED_WITH_DATA_SET (cca_client))
    {
      err = client_store_open (client, project_get_client_config_dir (project, client));
      if (err)
        fprintf (stderr, "%s: could not open client's store; not restoring its data set\n", __FUNCTION__);
      else
        project_restore_data_set (project, client);
    }
  
  if (CCA_CLIENT_CONFIGURED_WITH_DATA_SET (cca_client) ||
      CCA_CLIENT_CONFIGURED_WITH_FILE (cca_client))
    {
      if (client_get_name (client))
        {
          char * name;
          name = cca_strdup (client_get_name (client));
          client_set_name (client, NULL);
          project_name_client (project, client, name);
          free (name);
        }
      
    }
  
  project->clients = cca_list_append (project->clients, client);
}

void
project_add_client (project_t * project, client_t * client)
{
  cca_list_t * list;
  client_t * lost_client;
  int no_client_id;
  uuid_t id, lost_id;
  cca_client_t * cca_client, * lost_cca_client;
  
  
  cca_client = client_get_cca_client (client);
  
  if (CCA_CLIENT_NO_AUTORESUME (cca_client))
    {
      project_new_client (project, client);
      return;
    }
  
  cca_client_get_id (cca_client, id);
  no_client_id = uuid_is_null (id);
  uuid_clear (lost_id);
  
  CCA_DEBUGARGS ("attempting to resume client on connection %ld", client_get_conn_id (client));
  
  /* try and find a client we can resume */
  list = project->lost_clients;
  while (list)
    {
      lost_client = (client_t *) list->data;
      
      lost_cca_client = client_get_cca_client (lost_client);
      cca_client_get_id (lost_cca_client, lost_id);      

#ifdef LADCCA_DEBUG
      {      
        char * lost_str;
        lost_str = cca_strdup (cca_client_get_id_str (lost_cca_client));
        CCA_DEBUGARGS ("checking client '%s','%s' against lost client '%s','%s'",
                       cca_client_get_id_str (cca_client),
                       cca_client_get_class (cca_client),
                       lost_str,
                       cca_client_get_class (lost_cca_client));
        free (lost_str);
      }
#endif /* LADCCA_DEBUG */

      if ( (no_client_id && strcmp (cca_client_get_class (cca_client),
                                    cca_client_get_class (lost_cca_client)) == 0)
           || uuid_compare (id, lost_id) == 0)
        {
          CCA_DEBUGARGS ("resuming client '%s' of class '%s' with client on connection %ld",
                         cca_client_get_id_str (lost_cca_client),
                         cca_client_get_class (lost_cca_client),
                         client_get_conn_id (client));
          project_resume_client (project, client, lost_client);
          return;
        }
      
      list = list->next;
    }
  
  CCA_DEBUGARGS ("could not resume client on connection %ld", client_get_conn_id (client));
  project_new_client (project, client);
}

void
project_save_clients (project_t * project)
{
  cca_list_t * list;
  client_t * client;
  cca_client_t * cca_client;
  cca_event_t * event;
  
  for (list = project->clients; list; list = cca_list_next (list))
    {
      client = (client_t *) list->data;
      cca_client = client_get_cca_client (client);
      
      if (CCA_CLIENT_CONFIGURED_WITH_FILE (cca_client))
        {
          event = cca_event_new_with_type (CCA_Save_File);
          cca_event_set_string (event, project_get_client_id_dir (project, client));
          conn_mgr_send_client_cca_event (project->server->conn_mgr,
                                          client_get_conn_id (client),
                                          event);
        }
      
      if (CCA_CLIENT_CONFIGURED_WITH_DATA_SET (cca_client));
        {
          event = cca_event_new_with_type (CCA_Save_Data_Set);
          conn_mgr_send_client_cca_event (project->server->conn_mgr,
                                          client_get_conn_id (client),
                                          event);
        }
      
    }
}

int
project_write_client_jack_patches (FILE * file, project_t * project, client_t * client)
{
  unsigned long patch_count = 0;
  cca_list_t * list, * patches;
  jack_patch_t * patch;
  int err;
  
  patches = jack_mgr_get_client_patches (project->server->jack_mgr, client->cca_client->id);

  for (list = patches; list; list = cca_list_next (list))
    patch_count++;
  
  err = fprintf (file, "%ld\n", patch_count);
  if (err < 0)
    return 1;
  
  for (list = patches; list; list = cca_list_next (list))
    {
      patch = (jack_patch_t *) list->data;
      
      err = fprintf (file, "%s\n", jack_patch_get_spec (patch));
      if (err < 0)
        {
          for (; list; list = list->next)
            jack_patch_destroy ((jack_patch_t *) list->data);
          cca_list_free (patches);
          return 1;
        }

      jack_patch_destroy (patch);
    }

  cca_list_free (patches);
    
  return 0;
}

int
project_write_client_alsa_patches (FILE * file, project_t * project, client_t * client)
{
  unsigned long patch_count = 0;
  cca_list_t * list, * patches;
  alsa_patch_t * patch;
  int err;
  
  patches = alsa_mgr_get_client_patches (project->server->alsa_mgr, client->cca_client->id);
  for (list = patches; list; list = cca_list_next (list))
    patch_count++;
  
  err = fprintf (file, "%ld\n", patch_count);
  if (err < 0)
    return 1;
  
  for (list = patches; list; list = cca_list_next (list))
    {
      patch = (alsa_patch_t *) list->data;
      
      err = fprintf (file, "%s\n", alsa_patch_get_spec (patch));

      if (err < 0)
        {
          while (list)
            {
              alsa_patch_destroy (patch);
              list = list->next;
            }
          cca_list_free (client->alsa_patches);
          client->alsa_patches = NULL;
          return 1;
        }
      
      alsa_patch_destroy (patch);
      
    }
  cca_list_free (patches);

  return 0;
}

int
project_write_client_info (FILE * file, project_t * project, client_t * client)
{
  int err;
  int argc, i;
  char ** argv;
  cca_client_t * cca_client;
  
  cca_client = client_get_cca_client (client);
  
  err = fprintf (file, "%s\n", cca_client_get_class (cca_client));
  if (err < 0)
    return 1;

  err = fprintf (file, "%d\n", cca_client_get_flags (cca_client) & ~CCA_No_Autoresume);
  if (err < 0)
    return 1;

  err = fprintf (file, "%s\n", cca_client_get_id_str (cca_client));
  if (err < 0)
    return 1;
    
  err = fprintf (file, "%s\n", cca_client_get_working_dir (cca_client));
  if (err < 0)
    return 1;
  
  argc = cca_client_get_argc (cca_client);
  err = fprintf (file, "%d\n", argc);
  if (err < 0)
    return 1;
  
  argv = cca_client_get_argv (cca_client);
  for (i = 0; i < argc; i++)
    {
      err = fprintf (file, "%s\n", argv[i]);
      if (err < 0)
        return 1;
    }
  
  if (CCA_CLIENT_USES_JACK (cca_client))
    {
      err = project_write_client_jack_patches (file, project, client);
      if (err)
        return err;
    }

  if (CCA_CLIENT_USES_ALSA (cca_client))
    {
      err = project_write_client_alsa_patches (file, project, client);
      if (err)
        return err;
    }
  
  
  return 0;
}

int
project_write_info (project_t * project)
{
  const char * info_file_name;
  FILE * info_file;
  int err;
  unsigned long client_count = 0;
  cca_list_t * list;
  client_t * client;
  
  info_file_name = cca_get_fqn (project_get_directory (project), INFO_FILE);
  
  info_file = fopen (info_file_name, "w");
  if (!info_file)
    {
      fprintf (stderr, "%s: could not open info file '%s': %s\n",
               __FUNCTION__, info_file_name, strerror (errno));
      return 1;
    }
  
  /*
   * project name
   */
  err = fprintf (info_file, "%s\n", project_get_name (project));
  if (err < 0)
    {
      fprintf (stderr, "%s: could not write data to info file '%s': %s",
               __FUNCTION__, info_file_name, strerror (errno));
      fclose (info_file);
      return 1;
    }
  
  /*
   * client_count
   */
  list = project->clients;
  while (list)
    {
      client_count++;
      list = list->next;
    }
  err = fprintf (info_file, "%ld\n", client_count);
  if (err < 0)
    {
      fprintf (stderr, "%s: could not write data to info file '%s': %s",
               __FUNCTION__, info_file_name, strerror (errno));
      fclose (info_file);
      return 1;
    }
  
  /*
   * clients
   */
  list = project->clients;
  while (list)
    {
      client = (client_t *) list->data;
      
      err = project_write_client_info (info_file, project, client);
      if (err)
        {
          fprintf (stderr, "%s: could not write data to info file '%s': %s",
                   __FUNCTION__, info_file_name, strerror (errno));
          fclose (info_file);
          return 1;
        }
      
      list = list->next;
    }
  
  err = fclose (info_file);
  if (err == EOF)
    {
      fprintf (stderr, "%s: could not close info file '%s': %s",
               __FUNCTION__, info_file_name, strerror (errno));
      return 1;
    }
  
  return 0;
}

void
project_clear_lost_clients (project_t * project)
{
  cca_list_t * list;
  client_t * client;
  const char * client_dir;
  
  for (list = project->lost_clients; list; list = cca_list_next (list))
    {
      client = (client_t *) list->data;
      
      client_dir = project_get_client_id_dir (project, client);
      if (cca_dir_exists (client_dir))
        cca_remove_dir (client_dir);
      
      client_destroy (client);
    }
    
  cca_list_free (project->lost_clients);
  project->lost_clients = NULL;
}

void
project_save (project_t * project)
{
  int err;
  
  project_save_clients (project);
  
  err = project_write_info (project);
  if (err)
    {
      fprintf (stderr, "%s: error writing info file for project '%s'; aborting save\n",
               __FUNCTION__, project_get_name (project));
      return;
    }
  
  project_clear_lost_clients (project);
}

size_t
getlinenn (char **lineptr, size_t *n, FILE *stream)
{
  size_t size;
  char * ptr;
  
  size = getline (lineptr, n, stream);
  
  ptr = strchr (*lineptr, '\n');
  if (ptr)
    *ptr = '\0';
  
  return size;
}

jack_patch_t *
client_read_jack_patch (FILE * info_file)
{
  char * line = NULL;
  size_t line_size = 0;
  size_t err;
  jack_patch_t * patch;
  
  err = getlinenn (&line, &line_size, info_file);
  if (err == -1)
    return NULL;
  
  patch = jack_patch_new ();
  jack_patch_set_spec (patch, line);

  free (line);
  
  return patch;
}

alsa_patch_t *
client_read_alsa_patch (FILE * info_file)
{
  char * line = NULL;
  size_t line_size = 0;
  size_t err;
  alsa_patch_t * patch;
  
  err = getlinenn (&line, &line_size, info_file);
  if (err == -1)
    return NULL;
  
  patch = alsa_patch_new ();
  alsa_patch_set_spec (patch, line);

  free (line);
  
  return patch;
}

client_t *
client_read_info (FILE * info_file)
{
  client_t * client;
  cca_client_t * cca_client;
  int i, ierr;
  char * line = NULL;
  size_t line_size = 0;
  uuid_t id;
  int argc;
  char ** argv;
  unsigned long patch_count;
  jack_patch_t * jack_patch;
  alsa_patch_t * alsa_patch;
  size_t err;
  char * ptr;
  
  /* client class */
  err = getlinenn (&line, &line_size, info_file);
  if (err == -1)
    return NULL;
  
  client = client_new ();
  cca_client = cca_client_new ();
  client_set_cca_client (client, cca_client);
  cca_client_set_class (cca_client, line);
  
  /* client flags */
  err = getlinenn (&line, &line_size, info_file);
  if (err == -1)
    {
      client_destroy (client);
      return NULL;
    }
  
  cca_client_set_flags (cca_client, atoi (line));
  
  /* client id */
  err = getlinenn (&line, &line_size, info_file);
  if (err == -1)
    {
      client_destroy (client);
      return NULL;
    }
  
  ierr = uuid_parse (line, id);
  if (ierr == -1)
    {
      client_destroy (client);
      return NULL;
    }
  cca_client_set_id (cca_client, id);
  
  /* working dir */
  err = getlinenn (&line, &line_size, info_file);
  if (err == -1)
    {
      client_destroy (client);
      return NULL;
    }
  
  cca_client_set_working_dir (cca_client, line);
  
  
  /* argc */
  err = getlinenn (&line, &line_size, info_file);
  if (err == -1)
    {
      client_destroy (client);
      return NULL;
    }
  
  argc = atoi (line);
  
  /* argvs */
  argv = cca_malloc (sizeof (char *) * argc);
  for (i = 0; i < argc; i++)
    {
      err = getlinenn (&line, &line_size, info_file);
      if (err == -1)
        {
          int j;
          for (j = 0; j < i; j++)
            free (argv[j]);
          free (argv);
          client_destroy (client);
          return NULL;
        }
      
      argv[i] = cca_strdup (line);
    }
  
  cca_client_set_args (cca_client, argc, argv);
/*  for (i = 0; i < argc; i++)
    {
      free (argv[i]);
    }
  free (argv); */
  
  /*
   * jack patches
   */
  if (CCA_CLIENT_USES_JACK (cca_client))
    {
      err = getlinenn (&line, &line_size, info_file);
      if (err == -1)
        {
          client_destroy (client);
          return NULL;
        }
  
      patch_count = strtoul (line, &ptr, 10);
      free (line);
      if (ptr == line ||
          patch_count == ULONG_MAX)
        {
          client_destroy (client);
          return NULL;
        }
  
      for (i = 0; i < patch_count; i++)
        {
          jack_patch = client_read_jack_patch (info_file);
          if (!jack_patch)
            {
              client_destroy (client);
              return NULL;
            }
      
          client->jack_patches = cca_list_append (client->jack_patches, jack_patch);
        }
    }
    
    
  /*
   * alsa patches
   */
  if (CCA_CLIENT_USES_ALSA (cca_client))
    {
      line = NULL;
      line_size = 0;
      err = getlinenn (&line, &line_size, info_file);
      if (err == -1)
        {
          client_destroy (client);
          return NULL;
        }
  
      patch_count = strtoul (line, &ptr, 10);
      free (line);
      if (ptr == line ||
          patch_count == ULONG_MAX)
        {
          client_destroy (client);
          return NULL;
        }
  
      for (i = 0; i < patch_count; i++)
        {
          alsa_patch = client_read_alsa_patch (info_file);
          if (!alsa_patch)
            {
              client_destroy (client);
              return NULL;
            }
      
          client->alsa_patches = cca_list_append (client->alsa_patches, alsa_patch);
        }
    }
  
  return client;
}

project_t *
project_read_info (server_t * server, FILE * info_file)
{
  project_t * project;
  char * line = NULL;
  size_t line_size = 0;
  size_t err;
  unsigned long client_count, i;
  client_t * client;
  
  
  /* project name */
  err = getlinenn (&line, &line_size, info_file);
  if (err == -1)
    return NULL;
   
  project = project_new (server);
  project_set_name (project, line);

  
  /* client count */
  err = getlinenn (&line, &line_size, info_file);
  if (err == -1)
    {
      project_destroy (project);
      return NULL;
    }
  
  client_count = strtoul (line, NULL, 10);
  
  free (line);
  
  /* clients */
  for (i = 0; i < client_count; i++)
    {
      client = client_read_info (info_file);
      if (!client)
        {
          project_destroy (project);
          return NULL;
        }
      
      cca_client_set_project_name (client->cca_client, project_get_name (project));
      
      project->lost_clients = cca_list_append (project->lost_clients, client);
    }
  
  return project;
}

project_t *
project_restore (server_t * server, const char * dir)
{
  cca_list_t * list;
  project_t * project;
  const char * filename;
  FILE * info_file;
  int err;
  
  CCA_DEBUGARGS ("attempting to restore project in dir '%s'", dir);
  
  /* check if we've already got it open */
  list = server->projects;
  while (list)
    {
      project = (project_t *) list->data;
      
      if (strcmp (project_get_directory (project), dir) == 0)
        {
          fprintf (stderr, "%s: cannot restore project from directory '%s': a project, '%s', is already active with this directory\n",
                   __FUNCTION__, dir, project_get_name (project));
          return NULL;
        }
      
      list = list->next;
    }
  
  filename = cca_get_fqn (dir, INFO_FILE);
  
  info_file = fopen (filename, "r");
  if (!info_file)
    {
      fprintf (stderr, "%s: could not open info file '%s': %s\n",
               __FUNCTION__, filename, strerror (errno));
      return NULL;
    }
  
  project = project_read_info (server, info_file);
  if (!project)
    {
      fprintf (stderr, "%s: could not read project from info file '%s': %s\n",
               __FUNCTION__, filename, strerror (errno));
      fclose (info_file);
      return NULL;
    }
    
  project_set_directory (project, dir);
  
  err = fclose (info_file);
  if (err == -1)
    {
      fprintf (stderr, "%s: could not close project info file '%s': %s",
               __FUNCTION__, filename, strerror (errno));
    }

#ifdef LADCCA_DEBUG
  {
    cca_list_t * client_list;
    client_t * client;
    cca_list_t * list;
    
  
    CCA_DEBUG     ("resored project with:");
    CCA_DEBUGARGS ("  directory: '%s'", project_get_directory (project));
    CCA_DEBUGARGS ("  name:      '%s'", project_get_name (project));
    CCA_DEBUG     ("  clients:");
    client_list = project->lost_clients;
    while (client_list)
      {
        client = (client_t *) client_list->data;
        
        CCA_DEBUG     ("  ------");
        CCA_DEBUGARGS ("    id:          '%s'", cca_client_get_id_str (client->cca_client));
        CCA_DEBUGARGS ("    name:        '%s'", client_get_name (client));
        CCA_DEBUGARGS ("    working dir: '%s'", cca_client_get_working_dir (client->cca_client));
        CCA_DEBUGARGS ("    flags:       %d",   cca_client_get_flags (client->cca_client));
        if (client->alsa_patches)
          {
            alsa_patch_t * patch;
            
            CCA_DEBUG ("    alsa patches:");
            list = client->alsa_patches;
            while (list)
              {
                patch = (alsa_patch_t *) list->data;
                
                CCA_DEBUGARGS ("      %s", alsa_patch_get_spec (patch));
                
                list = list->next;
              }
          }
        else
          CCA_DEBUG ("    no alsa patches");

        if (client->jack_patches)
          {
            jack_patch_t * patch;
            
            CCA_DEBUG ("    jack patches:");
            list = client->jack_patches;
            while (list)
              {
                patch = (jack_patch_t *) list->data;
                
                CCA_DEBUGARGS ("      %s", jack_patch_get_spec (patch));
                
                list = list->next;
              }
          }
        else
          CCA_DEBUG ("    no jack patches");
        
        client_list = client_list->next;
      }
  }
#endif
  
  return project;
}

void
project_lose_client (project_t * project, client_t * client,
                     cca_list_t * jack_patches, cca_list_t * alsa_patches)
{
  cca_client_t * cca_client;
  cca_client = client_get_cca_client (client);
  
  CCA_DEBUGARGS ("losing client '%s' from project '%s'",
                 cca_client_get_id_str (cca_client),
                 project_get_name (project));
  project->clients = cca_list_remove (project->clients, client);

  if (CCA_CLIENT_CONFIGURED_WITH_DATA_SET (cca_client))
    {
      store_t * store;
      store = client_get_store (client);
      
      if (store_get_keys (store))
        store_write (store);
      else
        if (cca_dir_exists (store_get_dir (store)))
          cca_remove_dir (store_get_dir (store));
    }
  

  if (CCA_CLIENT_CONFIGURED_WITH_DATA_SET (cca_client) ||
      CCA_CLIENT_CONFIGURED_WITH_FILE (cca_client))
    {
      const char * dir;
      dir = project_get_client_id_dir (project, client);
      
      if (cca_dir_exists (dir) && cca_dir_empty (dir))
        cca_remove_dir (dir);
      
      dir = cca_get_fqn (project_get_directory (project), ID_DIR);
      if (cca_dir_exists (dir) && cca_dir_empty (dir))
        cca_remove_dir (dir);
    
      if (client_get_name (client))
        unlink (project_get_client_name_dir (project, client));
      
    }
  
  client->jack_patches = cca_list_concat (client->jack_patches, jack_patches);
  client->alsa_patches = cca_list_concat (client->alsa_patches, alsa_patches);
  project->lost_clients = cca_list_append (project->lost_clients, client);
}

/* EOF */

