/*
 *   LADCCA
 *    
 *   Copyright (C) 2002 Robert Ham (node@users.sourceforge.net)
 *    
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define _GNU_SOURCE

#include <signal.h>
#include <assert.h>

#include <ladcca/internal_headers.h>

#include "server.h"
#include "project.h"
#include "client_event.h"
#include "loader.h"

void server_create_loader (server_t * server);
void server_set_loader    (server_t * server, loader_t * loader);
void server_set_ui_project (server_t * server, const char * project);
void server_set_default_dir (server_t * server, const char * default_dir);


void
server_init (server_t * server)
{
  server->conn_mgr      = NULL;
  server->jack_mgr      = NULL;
  server->alsa_mgr      = NULL;
  server->loader        = NULL;
  server->loader_quit   = 0;
  server->projects      = NULL;
  server->server_events = NULL;
  server->default_dir   = NULL;
  server->quit          = 0;
  
  pthread_mutex_init (&server->server_events_lock, NULL);
  pthread_cond_init (&server->server_event_cond, NULL);
}

void
server_free (server_t * server)
{
  server_set_ui_project  (server, NULL);
  server_set_default_dir (server, NULL);
}

server_t *
server_new (unsigned int listen_port, const char * default_dir)
{
  server_t * server;
  server = cca_malloc (sizeof (server_t));
  server_init (server);
  
  server_set_default_dir (server, default_dir);
  
  server_create_loader (server);
  
  server->jack_mgr = jack_mgr_new (server);
  server->alsa_mgr = alsa_mgr_new (server);
  server->conn_mgr = conn_mgr_new (server, listen_port);
  
  return server;
}

void
server_destroy (server_t * server)
{

  
  conn_mgr_destroy (server->conn_mgr);
  alsa_mgr_destroy (server->alsa_mgr);
  jack_mgr_destroy (server->jack_mgr);

  signal (SIGCHLD, SIG_IGN);
  loader_close (server->loader);
  loader_destroy (server->loader);

  server_free (server);
  free (server);
}

void
server_create_loader  (server_t * server)
{
  loader_t * loader;
  int err;
  
  if (server->loader)
    {
      loader_close (server->loader);
      loader_destroy (server->loader);
      server->loader = NULL;
      server->loader_quit = 0;
    }
  
  loader = loader_create (1);
  if (!loader)
    {
      fprintf (stderr, "%s: could not create loader; aborting\n", __FUNCTION__);
      abort ();
    }
  
  err = loader_fork (loader);
  if (err)
    {
      fprintf (stderr, "%s: could not fork loader; aborting\n", __FUNCTION__);
      abort ();
    }
    
  server_set_loader (server, loader);
}

const char *
server_get_ui_project (server_t * server)
{
  return server->ui_project;
}


void
server_set_ui_project (server_t * server, const char * project)
{
  set_string_property (server->ui_project, project);
}

void
server_set_loader (server_t * server, loader_t * loader)
{
  server->loader = loader;
}

conn_mgr_t *
server_get_conn_mgr (server_t * server)
{
  return server->conn_mgr;
}

void
server_set_default_dir (server_t * server, const char * default_dir)
{
  set_string_property (server->default_dir, default_dir);
}

const char *
server_get_default_dir (server_t * server)
{
  return server->default_dir;
}

int
server_get_loader_quit (server_t * server)
{
  return server->loader_quit;
}

void
server_set_loader_quit (server_t * server, int quit)
{
  server->loader_quit = quit;
}

void
server_send_event (server_t * server, server_event_t * server_event)
{
  CCA_DEBUGARGS ("sending server event with type %d", server_event_get_type (server_event));
  pthread_mutex_lock (&server->server_events_lock);
  server->server_events = cca_list_append (server->server_events, server_event);
  pthread_mutex_unlock (&server->server_events_lock);
  pthread_cond_signal (&server->server_event_cond);
}


/*****************************
 ******* server stuff ********
 *****************************/

project_t *
server_find_project_by_name (server_t * server, const char * project_name)
{
  cca_list_t * list;
  project_t * project;
  
  list = server->projects;
  while (list)
    {
      project = (project_t *) list->data;
      
      if (strcmp (project_get_name (project), project_name) == 0)
        return project;
      
      list = list->next;
    }
  
  return NULL;
}

project_t *
server_find_project_by_conn_id (server_t * server, unsigned long conn_id)
{
  cca_list_t * list, * client_list;
  project_t * project;
  client_t * client;
  
  list = server->projects;
  while (list)
    {
      project = (project_t *) list->data;
      
      client_list = project->clients;
      while (client_list)
        {
          client = (client_t *) client_list->data;
          
          if (client_get_conn_id (client) == conn_id)
            return project;
          
          client_list = client_list->next;
        }
      
      list = list->next;
    }
  
  return NULL;
}

void
server_find_project_and_client_by_conn_id (server_t * server, unsigned long conn_id,
                                           project_t ** project_ptr, client_t ** client_ptr)
{
  cca_list_t * list, * client_list;
  project_t * project;
  client_t * client;
  
  list = server->projects;
  while (list)
    {
      project = (project_t *) list->data;
      
      client_list = project->clients;
      while (client_list)
        {
          client = (client_t *) client_list->data;
          
          if (client_get_conn_id (client) == conn_id)
            {
              if (client_ptr)
                *client_ptr = client;
                
              if (project_ptr)
                *project_ptr = project;
                
              return;
            }
          
          client_list = client_list->next;
        }
      
      list = list->next;
    }
  
  if (client_ptr)
    *client_ptr = NULL;
    
  if (project_ptr)
    *project_ptr = NULL;
}

client_t *
server_find_client_by_conn_id (server_t * server, unsigned long conn_id)
{
  cca_list_t * list;
  project_t * project;
  client_t * client;
  
  project = server_find_project_by_conn_id (server, conn_id);
  
  if (!project)
    return NULL;
  
  list = project->clients;
  while (list)
    {
      client = (client_t *) list->data;
      
      if (client_get_conn_id (client) == conn_id)
        return client;
      
      list = list->next;
    }
  
  return NULL;
}

const char *
server_create_new_project_name (server_t * server)
{
  static char new_name[32];
  int num = 1;
  
  do
    {
      sprintf (new_name, "project-%d", num);
      num++;
    }
  while (project_name_exists (server->projects, new_name));
  
  return new_name;
}

project_t *
server_create_new_project (server_t * server, const char * name)
{
  project_t * project;
  char * dir;
  size_t dir_size;
  
  project = project_new (server);
  
  if (!name || !strlen (name))
    { /* create a project name */
      project_set_name (project, server_create_new_project_name (server));
    }
  else
    project_set_name (project, name);
  
  dir_size = strlen (server_get_default_dir (server)) + 1 +
             strlen (project_get_name (project)) + 1;
  dir = cca_malloc (dir_size);
  sprintf (dir, "%s/%s", server_get_default_dir (server), project_get_name (project));
  project_set_directory (project, dir);
  free (dir);
  
  CCA_DEBUGARGS ("created project '%s' with directory '%s'", project_get_name (project),
                 project_get_directory (project));
  
  server->projects = cca_list_append (server->projects, project);
  
  return project;
}

void
server_add_client (server_t * server, client_t * client)
{
  project_t * project;
  cca_client_t * cca_client;
  
  cca_client = client_get_cca_client (client);
  
  if (CCA_CLIENT_IS_SERVER_UI(cca_client))
    cca_client_set_project_name (cca_client, UI_PROJECT_NAME);
  
  project = server_find_project_by_name (server, cca_client_get_project_name(cca_client));
  if (!project)
    { /* try and find an existing project */
      if ((!cca_client_get_project_name (cca_client) ||
           !strlen (cca_client_get_project_name (cca_client)))
          && server->projects->next)
        {
          project = (project_t *) server->projects->next->data;
          cca_client_set_project_name (cca_client, project_get_name (project));
        }
      else
        project = server_create_new_project (server, cca_client_get_project_name(cca_client));
    }
  
  project_add_client (project, client);
  
}

void
server_event_client_connect (server_t * server, server_event_t * event)
{
  client_t * client;
  
  client = client_new ();
  client_set_conn_id (client, server_event_get_conn_id (event));
  client_set_cca_client (client, server_event_take_cca_client (event));
  
  server_add_client (server, client);
}

void
server_event_client_disconnect (server_t * server, server_event_t * event)
{
  client_t * client;
  project_t * project;
  cca_list_t * jack_patches = NULL;
  cca_list_t * alsa_patches = NULL;
  
  server_find_project_and_client_by_conn_id (server, server_event_get_conn_id (event), &project, &client);
  if (!client)
    {
      fprintf (stderr, "%s: recieved disconnect for connection id %ld, but no client exists for it :/\n",
               __FUNCTION__, server_event_get_conn_id (event));
      return;
    }
  
  if (CCA_CLIENT_USES_JACK (client->cca_client))
    jack_patches = jack_mgr_remove_client (server->jack_mgr, client->cca_client->id);
  
  if (CCA_CLIENT_USES_ALSA (client->cca_client))
    alsa_patches = alsa_mgr_remove_client (server->alsa_mgr, client->cca_client->id);
  
  project_lose_client (project, client, jack_patches, alsa_patches);
}

void
server_event_client_event (server_t * server, server_event_t * server_event)
{
  cca_event_t * event;
  project_t * project;
  client_t * client;
  
  event = server_event_take_cca_event (server_event);
  
  /* find the project and client */
  server_find_project_and_client_by_conn_id (server, server_event_get_conn_id (server_event), &project, &client);
  if (!client)
    {
      fprintf (stderr, "%s: recieved server event from connection manager for unknown client with connection id %ld\n",
               __FUNCTION__, server_event_get_conn_id (server_event));
      return;
    }

  switch (cca_event_get_type (event))
    {
    case CCA_Client_Name:
      server_cca_event_client_name (project, client, cca_event_get_string (event));
      break;
    case CCA_Jack_Client_Name:
      server_cca_event_jack_client_name (server, client, cca_event_get_string (event));
      break;
    case CCA_Alsa_Client_ID:
      server_cca_event_alsa_client_id (server, project, client, cca_event_get_string(event));
      break;
    case CCA_Restore_Data_Set:
      server_cca_event_restore_data_set (server, project, client);
      break;
    case CCA_Save_Data_Set:
      server_cca_event_save_data_set (client);
      break;
    case CCA_Save:
      server_cca_event_save (server, project, client);
      break;
      
    case CCA_Project:
      server_cca_event_project (server, project, client, cca_event_get_string (event));
      break;
    case CCA_Project_Dir:
      server_cca_event_project_dir (server, project, client, cca_event_get_string (event));
      break;
    case CCA_Project_Name:
      server_cca_event_project_name (server, project, client, cca_event_get_string (event));
      break;
    case CCA_Projects:
      server_cca_event_projects (server, client);
      break;
    case CCA_Restore:
      server_cca_event_restore (server, cca_event_get_string (event));
      break;
    
    default:
      fprintf (stderr, "%s: received unknown cca event of type %d\n", __FUNCTION__,
               cca_event_get_type (event));
      break;
    }
  
  cca_event_destroy (event);
}

void
server_event_client_config (server_t * server, server_event_t * event)
{
  client_t * client;
  store_t * store;

  /* find the project and client */
  server_find_project_and_client_by_conn_id (server, server_event_get_conn_id (event), NULL, &client);
  if (!client)
    {
      fprintf (stderr, "%s: recieved server event from connection manager for unknown client with connection id %ld\n",
               __FUNCTION__, server_event_get_conn_id (event));
      return;
    }
  
  if (!CCA_CLIENT_CONFIGURED_WITH_DATA_SET(client->cca_client))
   {
     fprintf (stderr, "%s: client '%s' sent us a config but doesn't describe itself as configured with a data set\n",
              __FUNCTION__, cca_client_get_id_str (client->cca_client));
     return;
   }
  
  store = client_get_store (client);
  
  store_set_config (store, server_event_get_cca_config (event));
}

void
server_deal_with_event (server_t * server, server_event_t * event)
{
  switch (server_event_get_type (event))
    {
    case Client_Connect:
      server_event_client_connect (server, event);
      break;
    case Client_Disconnect:
      server_event_client_disconnect (server, event);
      break;
    case Client_Event:
      server_event_client_event (server, event);
      break;
    case Client_Config:
      server_event_client_config (server, event);
      break;
    default:
      fprintf (stderr, "%s: recieved unknown server event of type %d\n",
               __FUNCTION__, server_event_get_type (event));
    }
  
  server_event_destroy (event);
}

void
server_create_ui_project (server_t * server)
{
  project_t * project;
  
  project = server_create_new_project (server, UI_PROJECT_NAME);
  project_set_directory (project, cca_get_fqn (getenv ("HOME"), UI_PROJECT_DIR));
  
  server_set_ui_project (server, UI_PROJECT_NAME);
}

void
server_main (server_t * server)
{
  server_event_t * server_event;
  cca_list_t * server_events, * list;
  struct timespec ts;
  struct timeval now;
  
  server_create_ui_project (server);
  
  while (!server->quit)
    {
      pthread_mutex_lock (&server->server_events_lock);
      server_events = server->server_events;
      if (!server_events)
        {
          gettimeofday(&now, NULL);
          ts.tv_sec = now.tv_sec + 1;
          ts.tv_nsec = now.tv_usec * 1000;
          
          pthread_cond_timedwait (&server->server_event_cond,
                                          &server->server_events_lock,
                                          &ts);

          server_events = server->server_events;
        }

      server->server_events = NULL;
      pthread_mutex_unlock (&server->server_events_lock);
      
      for (list = server_events; list; list = cca_list_next (list))
        {
          server_event = (server_event_t *) list->data;
          server_deal_with_event (server, server_event);
        }
      
      cca_list_free (server_events);
    }
  
  
  
  CCA_DEBUG ("finished");
}

/* EOF */

