/*
 *   LADCCA
 *    
 *   Copyright (C) 2002 Robert Ham (node@users.sourceforge.net)
 *    
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <pthread.h>
#include <sys/socket.h>
#include <netdb.h>
extern int h_errno;
#include <errno.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
                     

#include <ladcca/ladcca.h>
#include <ladcca/internal_headers.h>

cca_args_t *
cca_extract_args (int * argc, char *** argv)
{
  int i, j, valid_count;
  char * ptr;
  long l;
  cca_args_t * args;
  
  args = cca_args_new ();
  
  for (i = 1; i < *argc; ++i)
    {
      if (strncasecmp ("--ladcca-server=", (*argv)[i], 16) == 0)
        {
          CCA_DEBUGARGS ("setting server from command line: '%s'", (*argv)[i] + 16);
          cca_args_set_server (args, (*argv)[i] + 16);
          (*argv)[i] = NULL;
          continue;
        }

      if (strncasecmp ("--ladcca-port=", (*argv)[i], 14) == 0)
        {
          l = strtol ((*argv)[i] + 14, &ptr, 10);
          
          if (l == LONG_MAX ||
              l == LONG_MIN ||
              l <= 0 ||
              l > INT_MAX ||
              ptr == (*argv)[i] + 15 )
/*            comm->port = CCA_DEFAULT_PORT; */
            l++;
          else
            {
              CCA_DEBUGARGS ("setting port from command line: %ld", l);
              cca_args_set_port (args, l);
            }
          (*argv)[i] = NULL;
          continue;
        }
        
      if (strncasecmp ("--ladcca-project=", (*argv)[i], 17) == 0)
        {
          CCA_DEBUGARGS ("setting project from command line: '%s'", (*argv)[i] + 17);
          cca_args_set_project (args, (*argv)[i] + 17);
          (*argv)[i] = NULL;
          continue;
        }
        
      if (strncmp ("--ladcca-id=", (*argv)[i], 12) == 0)
        {
          uuid_t id;
          
          CCA_DEBUGARGS ("setting id from command line: '%s'", (*argv)[i] + 12);
          
          j = uuid_parse ((*argv)[i] + 12, id);
          
          CCA_DEBUG ("id parsed");
          
          (*argv)[i] = NULL;
          
          if (j == -1)
            {
              fprintf(stderr, "%s: ERROR PARSING ID FROM COMMAND LINE!  THIS IS BAD!\n", __FUNCTION__);
            }
          else
            {
              cca_args_set_id (args, id);
            }
          continue;
        }

      if (strncmp ("--ladcca-no-autoresume", (*argv)[i], 22) == 0)
        {
          CCA_DEBUG ("setting CCA_No_Autoresume flag from command line");
          
          cca_args_set_flag (args, CCA_No_Autoresume);
          
          (*argv)[i] = NULL;
          
          continue;
        }
    }
    
  CCA_DEBUG ("args checked");
  
  /* sort out the argv pointers */
  valid_count = *argc;
  for (i = 1; i < valid_count; ++i)
    {
      CCA_DEBUGARGS ("checking argv[%d]", i);
      if ((*argv)[i] == NULL)
        {
          
          for (j = i; j < *argc - 1; ++j)
            (*argv)[j] = (*argv)[j + 1];
            
          valid_count--;
          i--;
        }
    }

  CCA_DEBUG ("done");
  
  *argc = valid_count;
  
  cca_args_set_args (args, *argc, (const char * const *) *argv);
  
  return args;
}

static void
lookup_server_name (cca_client_t * client)
{
  struct sockaddr_in server_addr;
  size_t server_addr_size;
  struct hostent * server_entry;
  int err;
  
  server_addr_size = sizeof (server_addr);
  err = getpeername (client->comm->socket,
                     (struct sockaddr *) &server_addr,
                     &server_addr_size);
  if (err)
    {
      fprintf (stderr, "%s: could not get server's ip address: %s\n",
               __FUNCTION__, strerror (errno));
      cca_client_set_server_name (client, "unknown-server");
      return;
    }
  
  server_entry = gethostbyaddr (&server_addr.sin_addr, sizeof (server_addr.sin_addr), AF_INET);
  if (!server_entry)
    {
      fprintf (stderr, "%s: could not find server's host name: %s\n",
               __FUNCTION__, hstrerror (h_errno));
      cca_client_set_server_name (client, inet_ntoa (server_addr.sin_addr));
      return;
    }
  
  cca_client_set_server_name (client, server_entry->h_name);
}

cca_client_t *
cca_init (cca_args_t * args,
          const char * class,
          int client_flags,
          cca_protocol_t protocol)
{
  cca_comm_t * comm;
  cca_client_t * client;
  uuid_t id;
  int err;
  char * str;
  const char * cstr;
  int port;
  
  
  /* create the client comm */
  comm = cca_malloc (sizeof (cca_comm_t));
  
  uuid_clear (id);
  
  
  /* create the client */
  client = cca_malloc (sizeof (cca_client_t));
  cca_client_init (client);

  cca_client_set_comm (client, comm);
  cca_client_set_flags (client, client_flags | cca_args_get_flags (args));
  cca_client_set_class (client, class);
  client->argc = cca_args_get_argc (args);
  client->argv = cca_args_take_argv (args);
  
  str = getcwd (NULL, 0);
  if (!str)
    {
      fprintf (stderr, "%s: could not get current working directory: %s\n",
               __FUNCTION__, strerror (errno));
      str = getenv ("PWD");
      if (str)
        cca_client_set_working_dir (client, str);
      else
        cca_client_set_working_dir (client, getenv ("HOME"));
    }
  else
    {
      cca_client_set_working_dir (client, str);
      free (str);
    }

  cstr = cca_args_get_project (args);
  cca_client_set_project_name (client, cstr ? cstr : "");
  
  cca_args_get_id (args, id);
  cca_client_set_id (client, id);
  cca_client_set_server_connected (client, protocol);
  
  comm->client = client;
  
  
  /* try and connect to the server */
  CCA_DEBUG ("connecting to server");
  port  = cca_args_get_port (args);
  cstr = cca_args_get_server (args);
  err = cca_comm_connect_to_server (comm,
                                    cstr ? cstr : "localhost",
                                    port ? port : CCA_DEFAULT_PORT);
  if (err)
    {
      fprintf (stderr, "%s: could not connect to server '%s:%d' - disabling ladcca\n",
               __FUNCTION__, cstr ? cstr : "localhost", port ? port : CCA_DEFAULT_PORT);
      
      /* clean our shit up */
      free (client);
      free (comm);
      return NULL;
    }
  CCA_DEBUG ("connected to server");
  cca_client_set_server_connected (client, 1);
    
  /* set up the buffers */
  comm->events_in = NULL;
  pthread_mutex_init (&comm->events_in_lock, NULL);
  comm->configs_in = NULL;
  pthread_mutex_init (&comm->configs_in_lock, NULL);
  comm->comm_events_out = NULL;
  pthread_mutex_init (&comm->comm_events_out_lock, NULL);
  
  pthread_cond_init (&comm->send_conditional, NULL);
  
  comm->recv_close = 0;
  comm->send_close = 0;
  
  /* start the recv thread */
  err = pthread_create (&comm->recv_thread, NULL, cca_comm_recv_run, comm);
  if (err)
    {
      fprintf (stderr, "%s: error creating recieve thread - disabling ladcca: %s\n",
               __FUNCTION__, strerror (err));
               
      /* clean our shit up */
      pthread_mutex_destroy (&comm->configs_in_lock);
      pthread_mutex_destroy (&comm->events_in_lock);
      pthread_mutex_destroy (&comm->comm_events_out_lock);
      pthread_cond_destroy (&comm->send_conditional);
      free (client);
      free (comm);
      
      return NULL;
    }

  /* start the send thread */
  err = pthread_create (&comm->send_thread, NULL, cca_comm_send_run, comm);
  if (err)
    {
      fprintf (stderr, "%s: error creating send thread - disabling ladcca: %s\n",
               __FUNCTION__, strerror (err));
      
      comm->recv_close = 1;
      pthread_join (comm->recv_thread, NULL);
               
      /* clean our shit up */
      pthread_mutex_destroy (&comm->configs_in_lock);
      pthread_mutex_destroy (&comm->events_in_lock);
      pthread_mutex_destroy (&comm->comm_events_out_lock);
      pthread_cond_destroy (&comm->send_conditional);
      free (client);
      free (comm);
      
      return NULL;
    }
  
  /* look up the server name */
  lookup_server_name (client);
  
  return client;
}

unsigned int
cca_get_pending_event_count (cca_client_t * client)
{
  unsigned int count = 0;
  
  if (!cca_enabled (client))
    return 0;
  
  pthread_mutex_lock (&client->comm->events_in_lock);
  
  if (client->comm->events_in)
    count = cca_list_length (client->comm->events_in);
    
  pthread_mutex_unlock (&client->comm->events_in_lock);
  
  return count;
}

unsigned int
cca_get_pending_config_count (cca_client_t * client)
{
  unsigned int count = 0;
  
  if (!cca_enabled (client))
    return 0;
  
  pthread_mutex_lock (&client->comm->configs_in_lock);
  
  if (client->comm->events_in)
    count = cca_list_length (client->comm->configs_in);
    
  pthread_mutex_unlock (&client->comm->configs_in_lock);
  
  return count;
}

cca_event_t *
cca_get_event (cca_client_t * client)
{
  cca_event_t * event = NULL;
  
  if (!cca_enabled (client))
    return NULL;
  
  pthread_mutex_lock (&client->comm->events_in_lock);
  
  if (client->comm->events_in)
    {
      event = (cca_event_t *) client->comm->events_in->data;
      client->comm->events_in =
        cca_list_remove (client->comm->events_in, event);
    }
    
  pthread_mutex_unlock (&client->comm->events_in_lock);
  
  return event;
}

cca_config_t *
cca_get_config (cca_client_t * client)
{
  cca_config_t * config = NULL;
  
  if (!cca_enabled (client))
    return NULL;
  
  pthread_mutex_lock (&client->comm->configs_in_lock);
  
  if (client->comm->configs_in)
    {
      config = (cca_config_t *) client->comm->configs_in->data;
      client->comm->configs_in =
        cca_list_remove (client->comm->configs_in, config);
    }
    
  pthread_mutex_unlock (&client->comm->configs_in_lock);
  
  return config;
}

void
cca_send_comm_event (cca_comm_t * comm, cca_comm_event_t * event)
{
  pthread_mutex_lock (&comm->comm_events_out_lock);
  comm->comm_events_out = cca_list_append (comm->comm_events_out, event);
  pthread_mutex_unlock (&comm->comm_events_out_lock);
  pthread_cond_signal (&comm->send_conditional);
}

void
cca_send_event (cca_client_t * client, cca_event_t * event)
{
  cca_comm_event_t * comm_event;
  
  if (!cca_enabled (client))
    {
      cca_event_destroy (event);
      return;
    }
  
  comm_event = cca_malloc (sizeof (cca_comm_event_t));
  comm_event->type = CCA_Comm_Event_Event;
  comm_event->event_data.event = event;
  
  cca_send_comm_event (client->comm, comm_event);
}

void
cca_send_config (cca_client_t * client, cca_config_t * config)
{
  cca_comm_event_t * comm_event;
  
  if (!cca_enabled (client))
    {
      cca_config_destroy (config);
      return;
    }
  
  comm_event = cca_malloc (sizeof (cca_comm_event_t));
  comm_event->type = CCA_Comm_Event_Config;
  comm_event->event_data.config = config;
  
  cca_send_comm_event (client->comm, comm_event);
}

const char *
cca_get_server_name (cca_client_t * client)
{
  if (!cca_enabled (client))
    return "";
  
  return cca_client_get_server_name (client);
}

int
cca_server_connected (cca_client_t * client)
{
  if (!client)
    return 0;
  
  return client->server_connected;
}

void
cca_jack_client_name (cca_client_t * client, const char * name)
{
  cca_event_t * event;
  
  if (!cca_enabled (client))
    return;

  if (!name)
    return;

  event = cca_event_new_with_all (CCA_Jack_Client_Name, name);
  
  cca_send_event (client, event);
}

void
cca_alsa_client_id (cca_client_t * client, unsigned char id)
{
  cca_event_t * event;
  
  if (!cca_enabled (client))
    return;

  if (!id)
    return;

  event = cca_event_new ();
  cca_event_set_alsa_client_id (event, id);
  
  cca_send_event (client, event);
}

/* EOF */
